@@ -176,6 +176,12 @@ spi_object_append_interfaces (DBusMessageIter * iter, AtkObject * obj)
       dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &itf);
     }
 
+  if (atk_object_get_role (obj) == ATK_ROLE_APPLICATION)
+    {
+      itf = SPI_DBUS_INTERFACE_APPLICATION;
+      dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &itf);
+    }
+
   if (ATK_IS_COMPONENT (obj))
     {
       itf = SPI_DBUS_INTERFACE_COMPONENT;
@@ -227,14 +233,14 @@ spi_object_append_interfaces (DBusMessageIter * iter, AtkObject * obj)
 #if 0
   if (ATK_IS_STREAMABLE_CONTENT (obj))
     {
-      itf = "org.freedesktop.atspi.StreamableContent";
+      itf = "org.a11y.atspi.StreamableContent";
       dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &itf);
     }
 #endif
 
   if (ATK_IS_DOCUMENT (obj))
     {
-      itf = "org.freedesktop.atspi.Collection";
+      itf = "org.a11y.atspi.Collection";
       dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &itf);
       itf = SPI_DBUS_INTERFACE_DOCUMENT;
       dbus_message_iter_append_basic (iter, DBUS_TYPE_STRING, &itf);
