@@ -42,7 +42,7 @@ struct _DRouteContext
     DBusConnection       *bus;
     GPtrArray            *registered_paths;
 
-    gchar                *introspect_dir;
+    gchar                *introspect_string;
 };
 
 struct _DRoutePath
@@ -50,6 +50,7 @@ struct _DRoutePath
     DRouteContext        *cnx;
     GStringChunk         *chunks;
     GPtrArray            *interfaces;
+    GPtrArray            *introspection;
     GHashTable           *methods;
     GHashTable           *properties;
 
@@ -70,6 +71,9 @@ typedef struct PropertyPair
 static DBusHandlerResult
 handle_message (DBusConnection *bus, DBusMessage *message, void *user_data);
 
+static DBusMessage *
+droute_object_does_not_exist_error (DBusMessage *message);
+
 /*---------------------------------------------------------------------------*/
 
 static DRoutePath *
@@ -83,6 +87,7 @@ path_new (DRouteContext *cnx,
     new_path->cnx = cnx;
     new_path->chunks = g_string_chunk_new (CHUNKS_DEFAULT);
     new_path->interfaces = g_ptr_array_new ();
+    new_path->introspection = g_ptr_array_new ();
 
     new_path->methods = g_hash_table_new_full ((GHashFunc)str_pair_hash,
                                                str_pair_equal,
@@ -105,6 +110,7 @@ path_free (DRoutePath *path, gpointer user_data)
 {
     g_string_chunk_free  (path->chunks);
     g_ptr_array_free     (path->interfaces, TRUE);
+    g_ptr_array_free     (path->introspection, FALSE);
     g_hash_table_destroy (path->methods);
     g_hash_table_destroy (path->properties);
 }
@@ -121,14 +127,13 @@ path_get_datum (DRoutePath *path, const gchar *pathstr)
 /*---------------------------------------------------------------------------*/
 
 DRouteContext *
-droute_new (DBusConnection *bus, const char *introspect_dir)
+droute_new (DBusConnection *bus)
 {
     DRouteContext *cnx;
 
     cnx = g_new0 (DRouteContext, 1);
     cnx->bus = bus;
     cnx->registered_paths = g_ptr_array_new ();
-    cnx->introspect_dir = g_strdup(introspect_dir);
 
     return cnx;
 }
@@ -137,7 +142,6 @@ void
 droute_free (DRouteContext *cnx)
 {
     g_ptr_array_foreach (cnx->registered_paths, (GFunc) path_free, NULL);
-    g_free (cnx->introspect_dir);
     g_free (cnx);
 }
 
@@ -170,7 +174,10 @@ droute_add_one (DRouteContext *cnx,
 
     registered = dbus_connection_register_object_path (cnx->bus, path, &droute_vtable, new_path);
     if (!registered)
-        oom();
+      {
+        path_free (new_path, NULL);
+        return NULL;
+      }
 
     g_ptr_array_add (cnx->registered_paths, new_path);
     return new_path;
@@ -198,6 +205,7 @@ droute_add_many (DRouteContext *cnx,
 void
 droute_path_add_interface(DRoutePath *path,
                           const char *name,
+                          const char *introspect,
                           const DRouteMethod   *methods,
                           const DRouteProperty *properties)
 {
@@ -207,6 +215,7 @@ droute_path_add_interface(DRoutePath *path,
 
     itf = g_string_chunk_insert (path->chunks, name);
     g_ptr_array_add (path->interfaces, itf);
+    g_ptr_array_add (path->introspection, introspect);
 
     for (; methods != NULL && methods->name != NULL; methods++)
       {
@@ -250,7 +259,7 @@ impl_prop_GetAll (DBusMessage *message,
 
     void  *datum = path_get_datum (path, pathstr);
     if (!datum)
-	return NULL;
+	return droute_object_does_not_exist_error (message);
 
     dbus_error_init (&error);
     if (!dbus_message_get_args
@@ -321,7 +330,7 @@ impl_prop_GetSet (DBusMessage *message,
 
     datum = path_get_datum (path, pathstr);
     if (!datum)
-	return NULL;
+	return droute_object_does_not_exist_error (message);
 
     if (get && prop_funcs->get)
       {
@@ -400,35 +409,6 @@ static const char *introspection_node_element =
 static const char *introspection_footer =
 "</node>";
 
-static void
-append_interface (GString     *str,
-                  const gchar *interface,
-                  const gchar *directory)
-{
-    gchar *filename;
-    gchar *contents;
-    gsize len;
-
-    GError *err = NULL;
-
-    filename = g_build_filename (directory, interface, NULL);
-
-    if (g_file_get_contents (filename, &contents, &len, &err))
-      {
-        g_string_append_len (str, contents, len);
-      }
-    else
-      {
-        g_warning ("AT-SPI: Cannot find introspection XML file %s - %s",
-                   filename, err->message);
-        g_error_free (err);
-      }
-
-    g_string_append (str, "\n");
-    g_free (filename);
-    g_free (contents);
-}
-
 static DBusHandlerResult
 handle_introspection (DBusConnection *bus,
                       DBusMessage    *message,
@@ -452,11 +432,10 @@ handle_introspection (DBusConnection *bus,
 
     g_string_append_printf(output, introspection_node_element, pathstr);
 
-    for (i=0; i < path->interfaces->len; i++)
+    for (i=0; i < path->introspection->len; i++)
       {
-        gchar *interface = (gchar *) g_ptr_array_index (path->interfaces, i);
-        _DROUTE_DEBUG ("DRoute (appending interface): %s\n", interface);
-        append_interface(output, interface, path->cnx->introspect_dir);
+        gchar *introspect = (gchar *) g_ptr_array_index (path->introspection, i);
+        g_string_append (output, introspect);
       }
 
     g_string_append(output, introspection_footer);
@@ -497,15 +476,14 @@ handle_other (DBusConnection *bus,
 
     _DROUTE_DEBUG ("DRoute (handle other): %s|%s on %s\n", member, iface, pathstr);
 
-    datum = path_get_datum (path, pathstr);
-    if (!datum)
-	return result;
-
     func = (DRouteFunction) g_hash_table_lookup (path->methods, &pair);
     if (func != NULL)
       {
-
-        reply = (func) (bus, message, datum);
+        datum = path_get_datum (path, pathstr);
+        if (!datum)
+	    reply = droute_object_does_not_exist_error (message);
+        else
+            reply = (func) (bus, message, datum);
 
         if (!reply)
           {
@@ -536,21 +514,49 @@ handle_message (DBusConnection *bus, DBusMessage *message, void *user_data)
     const gint   type    = dbus_message_get_type (message);
     const gchar *pathstr = dbus_message_get_path (message);
 
+    DBusHandlerResult result = DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+
     _DROUTE_DEBUG ("DRoute (handle message): %s|%s of type %d on %s\n", member, iface, type, pathstr);
 
     /* Check for basic reasons not to handle */
     if (type   != DBUS_MESSAGE_TYPE_METHOD_CALL ||
         member == NULL ||
         iface  == NULL)
-        return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
+        return result;
 
     if (!strcmp (iface, "org.freedesktop.DBus.Properties"))
-        return handle_properties (bus, message, path, iface, member, pathstr);
+        result = handle_properties (bus, message, path, iface, member, pathstr);
+    else if (!strcmp (iface, "org.freedesktop.DBus.Introspectable"))
+        result = handle_introspection (bus, message, path, iface, member, pathstr);
+    else
+        result = handle_other (bus, message, path, iface, member, pathstr);
+#if 0
+    if (result == DBUS_HANDLER_RESULT_NOT_YET_HANDLED)
+        g_print ("DRoute | Unhandled message: %s|%s of type %d on %s\n", member, iface, type, pathstr);
+#endif
+      
+    return result;
+}
 
-    if (!strcmp (iface, "org.freedesktop.DBus.Introspectable"))
-        return handle_introspection (bus, message, path, iface, member, pathstr);
+/*---------------------------------------------------------------------------*/
 
-    return handle_other (bus, message, path, iface, member, pathstr);
+static DBusMessage *
+droute_object_does_not_exist_error (DBusMessage *message)
+{
+    DBusMessage *reply;
+    gchar       *errmsg;
+
+    errmsg= g_strdup_printf (
+            "Method \"%s\" with signature \"%s\" on interface \"%s\" could not be processed as object %s does not exist\n",
+            dbus_message_get_member (message),
+            dbus_message_get_signature (message),
+            dbus_message_get_interface (message),
+            dbus_message_get_path (message));
+    reply = dbus_message_new_error (message,
+                                    DBUS_ERROR_FAILED,
+                                    errmsg);
+    g_free (errmsg);
+    return reply;
 }
 
 /*---------------------------------------------------------------------------*/
